/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.se.eip.patterns;

import java.net.URI;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;

import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.transform.Source;
import javax.xml.xpath.XPathExpressionException;

import org.ow2.petals.component.framework.Constants.MEPConstants;
import org.ow2.petals.component.framework.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.util.Exchange;
import org.ow2.petals.component.framework.util.SourceHelper;
import org.ow2.petals.jbi.descriptor.Consumes;
import org.ow2.petals.se.eip.ExchangeContext;
import org.w3c.dom.Document;

/**
 * Receives a stream of messages and identifies messages that are correlated
 * (XPATH expression). Once a complete set of messages has been received (XPATH
 * expression on special message to decide that the set is complete), the
 * aggregator collects information from each correlated message and publishes a
 * single, aggregated message to the output channel for further processing.
 *
 * @author Adrien LOUIS - eBM WebSourcing
 * @author Christophe HAMERLING - eBM WebSourcing
 * @since 2.0
 *
 */
public class Aggregator extends AbstractPattern {

	private Map<String, List<Exchange>> exchanges;

	public static final String DEFAULT_ROOT_ELEMENT = "aggregate";

	public static final String CORRELATION_ID = "correlation";

	public static final String COMPLETE = "complete";

	public static final String ROOT_ELEMENT = "root-element";

	/*
	 * (non-Javadoc)
	 *
	 * @see org.ow2.petals.se.eip.patterns.AbstractPattern#process(org.ow2.petals.component.framework.util.Exchange,
	 *      org.ow2.petals.se.eip.ExchangeContext)
	 */
	public void process(Exchange exchange, ExchangeContext context)
			throws MessagingException {

		String correlationID = retrieveCorrelationId(exchange, context);
		if (correlationID == null) {
			throw new MessagingException(
					"No correlation ID found in the message");
		}

		boolean complete = isComplete(exchange, context);
		List<Exchange> exchanges = this.exchanges.get(correlationID);
		if (exchanges == null && !complete) {
			exchanges = new ArrayList<Exchange>();
			this.exchanges.put(correlationID, exchanges);
		}

		if (complete) {
			logger.info("Sequence is complete");
			List<Consumes> consumes = context.getSUConsumes(exchange
					.getEndpoint());
			sendAggregatedMessage(correlationID, context, consumes);
		} else {
			logger.info("Storing exchange " + exchange.getExchangeId()
					+ " for correlation " + correlationID);
			exchanges.add(exchange);
			logger.info("Sequence size is : " + exchanges.size());
		}
	}

	/**
	 * Get the correlation ID from the message with an XPath expression or from
	 * the exchange property.
	 *
	 * @return
	 */
	private String retrieveCorrelationId(Exchange exchange,
			ExchangeContext context) {
		// retrieve the correlation from the message from an XPATH expression
		String result = null;
		String expression = context.getExtensions().getValue(CORRELATION_ID);
		if (expression != null) {
			try {
				result = PatternHelper.evaluateXPathExpression(expression,
						exchange.getInMessageContentAsDocument(), context);
				logger.info("Evaluate condition: \"" + expression
						+ "\", result: \"" + result + "\".");
			} catch (XPathExpressionException e) {
				final String err = "Error in XPathExpression \"" + expression
						+ "\", cause: \"" + e.getCause().getMessage() + "\".";
				logger.log(Level.SEVERE, err);
			} catch (MessagingException e) {
				logger.log(Level.SEVERE, e.getMessage());
			}
		} else {
			// get the correlation from the JBI message property
			result = (String) exchange.getProperty(CORRELATION_ID);
		}
		return result;
	}

	/**
	 *
	 * @param exchange
	 * @return
	 */
	private boolean isComplete(Exchange exchange, ExchangeContext context) {
		boolean result = false;
		String condition = context.getExtensions().getValue(COMPLETE);
		try {
			Document inMessage = exchange.getInMessageContentAsDocument();
			result = PatternHelper.evaluateXPathCondition(condition, inMessage
					.getFirstChild(), context);
			logger.info("Evaluate condition: \"" + condition + "\", result: \""
					+ result + "\".");
		} catch (MessagingException e) {
			logger.log(Level.SEVERE, e.getMessage());
		} catch (XPathExpressionException e) {
			final String err = "Error in XPathExpression \"" + condition
					+ "\", cause: \"" + e.getCause().getMessage() + "\".";
			logger.log(Level.SEVERE, err);

		}
		return result;
	}

	/**
	 *
	 * @param correlationId
	 * @param context
	 * @param consumes
	 */
	protected void sendAggregatedMessage(String correlationId,
			ExchangeContext context, List<Consumes> consumes)
			throws MessagingException {

		// get and remove messages from cache
		List<Exchange> exchanges = this.exchanges.get(correlationId);
		if (exchanges == null) {
			throw new MessagingException("No sequence found for correlation "
					+ correlationId);
		}

		// aggregate the messages into a single one...
		Source content = null;
		try {
			content = buildContent(exchanges, context);
		} catch (PEtALSCDKException e1) {
			throw new MessagingException(e1);
		}

		// send the aggregated message
		for (Consumes consume : consumes) {
			sendExchange(consume, content, context);
		}

		// FIXME : When can we delete the messages?
		this.exchanges.remove(correlationId);
	}

	/**
	 * @param aggregat
	 * @param toAggregate
	 */
	private Source buildContent(List<Exchange> toAggregate,
			ExchangeContext context) throws PEtALSCDKException {
		String root = getRootElement(context);
		StringBuffer content = new StringBuffer();
		content.append('<');
		content.append(root);
		content.append('>');
		for (Exchange exchange : toAggregate) {
			NormalizedMessage message = exchange.getInMessage();
			content.append(SourceHelper.createString(message.getContent()));
		}
		content.append("</");
		content.append(root);
		content.append('>');

		return SourceHelper.createSource(content.toString());
	}

	/**
	 * Get the root element name of the aggregated message
	 *
	 * @param context
	 * @return
	 */
	private String getRootElement(ExchangeContext context) {
		return context.getExtensions().getValue(ROOT_ELEMENT,
				DEFAULT_ROOT_ELEMENT);
	}

	/**
	 *
	 * @param consumes
	 * @param aggregat
	 * @param context
	 * @throws MessagingException
	 */
	private void sendExchange(Consumes consumes, Source content,
			ExchangeContext context) throws MessagingException {
		// send the message to the JBI Service
		Exchange exchange = context.createConsumeExchange(consumes);
		exchange.setInMessageContent(content);
		context.sendSync(exchange);

		// TODO handle response...

	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.ow2.petals.se.eip.patterns.AbstractPattern#processMEPValidation(java.net.URI)
	 */
	protected boolean validateMEP(URI pattern) {
		return MEPConstants.IN_ONLY_PATTERN.equals(pattern)
				|| MEPConstants.ROBUST_IN_ONLY_PATTERN.equals(pattern);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.ow2.petals.se.eip.patterns.Pattern#init()
	 */
	public void init() {
		this.exchanges = new HashMap<String, List<Exchange>>();
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.ow2.petals.se.eip.patterns.AbstractPattern#getPatternName()
	 */
	@Override
	protected String getPatternName() {
		return "Aggregator";
	}
}
