/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.se.eip.patterns;

import java.net.URI;
import java.util.List;

import javax.jbi.JBIException;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;

import org.ow2.petals.component.framework.Constants.MEPConstants;
import org.ow2.petals.component.framework.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.util.Exchange;
import org.ow2.petals.component.framework.util.Extensions;
import org.ow2.petals.jbi.descriptor.Consumes;
import org.ow2.petals.se.eip.ExchangeContext;

/**
 * Change the MEP of an exchange. Use the MEP defined in the Consumes element.
 * <br>
 * Depending on the MEP, some response can be lost. TODO sendAsync not managed
 *
 * @author alouis, chamerling - eBM WebSourcing
 *
 */
public class Bridge extends AbstractPattern {

	/*
	 * (non-Javadoc)
	 * @see org.ow2.petals.se.eip.patterns.Pattern#processPattern(org.ow2.petals.component.framework.util.Exchange, org.ow2.petals.se.eip.ExchangeContext)
	 */
    public void process(Exchange exchange, ExchangeContext context) {
    	Exchange originalExchange = exchange;
        List<Consumes> consumesList = context.getSUConsumes(originalExchange
                .getEndpoint());

        if (consumesList.size() == 1) {
            try {
                Consumes consumes = consumesList.get(0);
                Extensions extensions = new Extensions(consumes.getExtensions());
                MEPConstants newMEP = PatternHelper.retrievePattern(extensions);

                NormalizedMessage inMessage = originalExchange.getInMessage();

                // create the exchange with the new MEP to send to the next
                // service
                Exchange newExchange = PatternHelper.createExchange(consumes,
                        newMEP.value(), inMessage, context);

                // send the exchange to the next service
                context.getLogger().info(
                        "original M.E.P. : " + originalExchange.getPattern());
                context.getLogger().info(
                        "new      M.E.P. : " + newExchange.getPattern());
                context.getLogger().info(
                        "send exchange to ...\n - itf="
                                + consumes.getInterfaceName() + "\n - srv="
                                + consumes.getServiceName() + "\n - ep ="
                                + consumes.getEndpointName());
                context.sendSync(newExchange);

                // process the MEP mapping
                processMEPMapping(originalExchange, newExchange, context);

                // close the forwarded exchange if active
                if (newExchange.isActiveStatus()) {
                    closeForwardedExchange(originalExchange, newExchange, context);
                }
            } catch (PEtALSCDKException e) {
                context.getLogger().severe(e.getMessage());
                originalExchange.setError(new Exception(e.getMessage()));
            } catch (JBIException e) {
                context.getLogger().severe(e.getMessage());
                originalExchange.setError(e);
            }
        }else{
        	String error="The related ServiceUnit definition must contains one, and only one, service to consume.";
        	context.getLogger().severe(error);
            originalExchange.setError(new Exception(error));
        }
    }

    /**
     * Process the mapping between the original and new MEP :
     * <li>If an error has been received, set it on the original MEP.</li>
     * <li>Depending on the original MEP, copy received Fault or Out message to
     * the original Exchange, if any; otherwise copy an empty Out message.</li>
     *
     * @param originalExchange
     * @param newExchange
     * @param context
     */
    protected void processMEPMapping(Exchange originalExchange,
            Exchange newExchange, ExchangeContext context) {

        // copy back the error from the forwarded exchange if any
        if (newExchange.isErrorStatus()) {
            setErrorOnOriginalExchange(originalExchange, newExchange, context);
        }

        // MEP mapping
        if (originalExchange.isInOnlyPattern()) {
            // InOnly
            // do nothing ...
        } else if (originalExchange.isRobustInOnlyPattern()) {
            //RobustInOnly
            // send back Fault if any
            if (newExchange.isActiveStatus() && newExchange.getFault() != null) {
                setFaultOnOriginalExchange(originalExchange, newExchange);
            }
        } else if (originalExchange.isInOptionalOutPattern()) {
            // InOptionalOut
            // send back Fault or Out if any
            if (newExchange.isActiveStatus() && newExchange.getFault() != null) {
                setFaultOnOriginalExchange(originalExchange, newExchange);
            } else if (newExchange.isActiveStatus()
                    && newExchange.getOutMessage() != null) {
                setOutOnOriginalExchange(originalExchange, newExchange);
            }
        } else if (originalExchange.isInOutPattern()) {
            // InOut
            // send back Fault or Out if any, otherwise return an empty Out
            if (newExchange.isActiveStatus() && newExchange.getFault() != null) {
                setFaultOnOriginalExchange(originalExchange, newExchange);
            } else if (newExchange.isActiveStatus()
                    && newExchange.getOutMessage() != null) {
                setOutOnOriginalExchange(originalExchange, newExchange);
            } else if (newExchange.isActiveStatus()) {
                setEmptyOutOnOriginalExchange(originalExchange);
            }
        }
    }

    /**
     * Set the original exchange to the Error of the forwarded exchange
     *
     * @param originalExchange
     * @param newExchange
     * @param context
     */
    protected void setErrorOnOriginalExchange(Exchange originalExchange,
            Exchange newExchange, ExchangeContext context) {
        context
                .getLogger()
                .info(
                        "set on the original exchange an error received from the forwarded exchange.");
        originalExchange.setError(newExchange.getError());
    }

    /**
     * copy on the original exchange the Fault of the forwarded exchange If the
     * fault can not be copied or set, set an error on the original exchange
     *
     * @param originalExchange
     * @param newExchange
     * @param context
     */
    protected void setOutOnOriginalExchange(Exchange originalExchange,
            Exchange newExchange) {
        logger.info(
                        "set on the original exchange an Out message received from the forwarded exchange.");
        try {
            NormalizedMessage out = originalExchange.getOutMessage();
            PatternHelper.copy(newExchange.getOutMessage(), out);
        } catch (MessagingException e) {
        	logger.severe(e.getMessage());
            originalExchange.setError(e);
        } catch (PEtALSCDKException e) {
        	logger.severe(e.getMessage());
            originalExchange.setError(e);
		}
    }

    /**
     * set a default message on the original exchange. set an error on the
     * original exchange if message can not be set
     *
     * @param originalExchange
     * @param context
     */
    protected void setEmptyOutOnOriginalExchange(Exchange originalExchange) {
        logger.info(
                "set on the original exchange an empty Out message.");
        try {
            originalExchange.setOutMessageContent("<success/>");
        } catch (MessagingException e) {
            logger.severe(e.getMessage());
            originalExchange.setError(e);
        }
    }

    /**
     * copy on the original exchange the Fault of the forwarded exchange If the
     * fault can not be copied or set, set an error on the original exchange
     *
     * @param originalExchange
     * @param newExchange
     * @param context
     */
    protected void setFaultOnOriginalExchange(Exchange originalExchange,
            Exchange newExchange) {
        logger.info(
                        "set on the original exchange a Fault received from the forwarded exchange.");
        try {
            Fault fault = originalExchange.createFault();
            PatternHelper.copy(newExchange.getFault(), fault);
            originalExchange.setFault(fault);
        } catch (MessagingException e) {
            logger.severe(e.getMessage());
            originalExchange.setError(e);
        } catch (PEtALSCDKException e) {
            logger.severe(e.getMessage());
            originalExchange.setError(e);
		}
    }

    /**
     * Close the forwarded exchange by sending a DONE status. If the status can
     * not be sent, set an error on the original exchange
     *
     * @param originalExchange
     * @param newExchange
     * @param context
     */
    protected void closeForwardedExchange(Exchange originalExchange,
            Exchange newExchange, ExchangeContext context) {
        logger.info("close the forwarded exchange.");
        try {
            newExchange.setDoneStatus();
            context.send(newExchange);
        } catch (MessagingException e) {
            logger.severe(e.getMessage());
            originalExchange.setError(e);
        }
    }

    /*
     * (non-Javadoc)
     * @see org.ow2.petals.se.eip.patterns.Pattern#init()
     */
    public void init() {
	}

    /*
     * (non-Javadoc)
     * @see org.ow2.petals.se.eip.patterns.AbstractPattern#getPatternName()
     */
	@Override
	protected String getPatternName() {
		return "Bridge";
	}

	/*
	 * (non-Javadoc)
	 * @see org.ow2.petals.se.eip.patterns.AbstractPattern#validateMEP(java.net.URI)
	 */
	@Override
	protected boolean validateMEP(URI mep) {
		return true;
	}

}
