/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.se.eip.patterns;

import java.net.URI;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;

import javax.jbi.JBIException;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;

import org.ow2.petals.component.framework.Constants.MEPConstants;
import org.ow2.petals.component.framework.util.Exchange;
import org.ow2.petals.component.framework.util.SourceHelper;
import org.ow2.petals.jbi.descriptor.Consumes;
import org.ow2.petals.se.eip.ExchangeContext;

/**
 * Aggregate all Consumes services defined in the service unit. <br>
 * Return the aggregated results.<br>
 * <br> - Accept InOut and InOptionalOut <br> - If a Fault or Error occurs on a
 * call, stop the process and return the Fault or Error Note : The exchange will
 * be sent back automatically by the framework
 * 
 * @author alouis, chamerling - eBM WebSourcing
 * 
 */
public class ScatterGather extends AbstractPattern {

	public static final String AGGREGATE_RESULT_TAG = "aggregate-result";

	/*
	 * (non-Javadoc)
	 * @see org.ow2.petals.se.eip.patterns.AbstractPattern#process(org.ow2.petals.component.framework.util.Exchange, org.ow2.petals.se.eip.ExchangeContext)
	 */
	public void process(Exchange exchange, ExchangeContext context) {
		Exchange originalExchange = exchange;

		List<Consumes> consumesList = context.getSUConsumes(originalExchange
				.getEndpoint());

		List<Exchange> sentExchanges = sendAsyncExchanges(consumesList,
				exchange, context);

		// when messages have been sent, wait for the responses from all
		// services
		if (sentExchanges != null) {
			String aggregateResult = receiveResponse(sentExchanges, exchange,
					context);

			// when all responses have be collected, send it back to the
			// consumer
			if (aggregateResult != null) {
				sendBackAggregateResult(aggregateResult, exchange);
			}
		}
	}

	/**
	 * Send a copy of the exchange's IN message to all the services to consume.
	 * Return the list of the Exchanges sent asynchronously. If an error occurs,
	 * set a Fault on the original exchange, and return null TODO if an exchange
	 * fails, close the previous sent exchange...
	 * 
	 * @param consumesList
	 * @param exchange
	 * @param context
	 * @return the list of the exchanges sent async, or null if a problem
	 *         occurred
	 */
	protected List<Exchange> sendAsyncExchanges(List<Consumes> consumesList,
			Exchange exchange, ExchangeContext context) {

		// assumes all services to aggregate are based on the IN-OUT MEP
		List<Exchange> sentExchanges = new ArrayList<Exchange>();
		try {
			NormalizedMessage inMessage = exchange.getInMessage();
			// send all exchanges asynchronously
			for (Consumes consumedService : consumesList) {
				Exchange tmpExchange = PatternHelper.createExchange(
						consumedService, MEPConstants.IN_OUT_PATTERN.value(),
						inMessage, context);
				logger.info(
						"send exchange to ...\n - itf="
								+ consumedService.getInterfaceName()
								+ "\n - srv="
								+ consumedService.getServiceName()
								+ "\n - ep ="
								+ consumedService.getEndpointName());
				context.sendAsync(tmpExchange);
				sentExchanges.add(tmpExchange);
			}
		} catch (JBIException e) {
			logger.log(Level.SEVERE, e.getMessage(), e);
			try {
				exchange.setFault(e);
			} catch (MessagingException ex) {
				logger.severe(ex.getMessage());
				exchange.setError(ex);
			}
			sentExchanges = null;
		}
		return sentExchanges;
	}

	/**
	 * 
	 * @param sentExchanges
	 * @param exchange
	 * @param context
	 * @return
	 */
	protected String receiveResponse(List<Exchange> sentExchanges,
			Exchange exchange, ExchangeContext context) {
		// wait for all exchanges responses
		// and build the aggregate result
		String aggregateResult = "<" + AGGREGATE_RESULT_TAG + ">";
		try {
			for (Exchange sentExchange : sentExchanges) {
				Exchange aggregateExchange = context.accept(sentExchange);
				// process the response. If fault or error, stop the
				// process
				// the 'processSentExchange()' has set the error or
				// fault
				// on the original exchange
				NormalizedMessage receivedMessage = PatternHelper
						.processSentExchange(aggregateExchange, exchange,
								context);
				if (receivedMessage != null) {
					aggregateResult += SourceHelper
							.createString(receivedMessage.getContent());
				} else {
					aggregateResult = null;
					break;
				}
			}
		} catch (JBIException e) {
			logger.log(Level.SEVERE, e.getMessage(), e);
			exchange.setError(e);
		} catch (InterruptedException e) {
			logger.log(Level.SEVERE, e.getMessage(), e);
			exchange.setError(e);
		}
		return aggregateResult;
	}

	/**
	 * Set the aggregateResult as the OUT response of the exchange. If it can
	 * not be set, set an error on the exchange
	 * 
	 * @param aggregateResult
	 * @param exchange
	 * @param context
	 */
	protected void sendBackAggregateResult(String aggregateResult,
			Exchange exchange) {

		logger.info("send response to the consumer.");
		aggregateResult += "</" + AGGREGATE_RESULT_TAG + ">";
		try {
			exchange.setOutMessageContent(aggregateResult);
		} catch (MessagingException e) {
			logger.log(Level.SEVERE, e.getMessage(), e);
			exchange.setError(e);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.ow2.petals.se.eip.patterns.Pattern#init()
	 */
	public void init() {
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.ow2.petals.se.eip.patterns.AbstractPattern#getPatternName()
	 */
	@Override
	protected String getPatternName() {
		return "ScatterGather";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.ow2.petals.se.eip.patterns.AbstractPattern#validateMEP(java.net.URI)
	 */
	@Override
	protected boolean validateMEP(URI mep) {
		return MEPConstants.IN_OUT_PATTERN.equals(mep)
				|| MEPConstants.IN_OPTIONAL_OUT_PATTERN.equals(mep);
	}
}
