/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.se.eip.patterns;

import java.net.URI;
import java.util.List;
import java.util.logging.Level;

import javax.jbi.JBIException;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;

import org.ow2.petals.component.framework.Constants.MEPConstants;
import org.ow2.petals.component.framework.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.util.Exchange;
import org.ow2.petals.jbi.descriptor.Consumes;
import org.ow2.petals.se.eip.ExchangeContext;

/**
 * Before sending the Exchange to the Service provider (or consumer for the
 * response), Copy the In or OUT message and send it to a "monitoring service" .
 * <br>
 * The message sent to the monitoring service is InOnly <br>
 * request mode : <br> - copy IN to the monitor and send the exchange to the
 * provider; response is received and send back to the consumer <br>
 * response mode : <br> - send the exchange to the provider, receive the
 * reponse, copy OUT/Fault to the monitor and send back the response to the
 * consumer <br>
 * request-response mode : <br>
 * copy in and out to the monitor
 * 
 * @author alouis, chamerling - eBM WebSourcing
 * 
 */
public class WireTap extends AbstractPattern {

	public static final String WAY = "way";
	public static final String REQUESTWAY = "request";
	public static final String RESPONSEWAY = "response";
	public static final String REQUESTRESPONSEWAY = "request-response";

	/*
	 * (non-Javadoc)
	 * @see org.ow2.petals.se.eip.patterns.Pattern#processPattern(org.ow2.petals.component.framework.util.Exchange, org.ow2.petals.se.eip.ExchangeContext)
	 */
	public void process(Exchange exchange, ExchangeContext context) {
		Exchange originalExchange = exchange;
		try {
			List<Consumes> consumesList = context
					.getSUConsumes(originalExchange.getEndpoint());
			int count = consumesList.size();
			if (count <= 0 || count > 2) {
				throw new PEtALSCDKException(
						"only 2 <Consumes> elements must be defined in the service-unit.");
			} else {
				sendExchange(consumesList, exchange, context);
			}
		} catch (JBIException e) {
			logger.log(Level.SEVERE, e.getMessage(), e);
			exchange.setError(e);
		}

	}

	/**
	 * Send the exchange to its destination and send a copy to the 'monitoring
	 * service'. <br>
	 * Depending on the "way" parameter , copy the IN or OUT( or fault) message
	 * and send it to the 'monitoring service', as InOnly exchange. <br>
	 * values for "way" parameter : <br>
	 * "request" send the IN message (InOnly MEP) (default mode) <br>
	 * "response" send the OUT or Fault or ERROR as an IN message (InOnly MEP)
	 * default mode for "way" is "request"
	 *
	 * @param consumesList
	 * @param exchange
	 * @param context
	 */
	protected void sendExchange(List<Consumes> consumesList, Exchange exchange,
			ExchangeContext context) {

		try {

			Consumes provider = consumesList.get(0);
			Consumes monitor = consumesList.get(1);

			// send IN to the monitor if Request way is defined
			sendRequestCopyIfNeeded(monitor, exchange, context);

			// send (sync) exchange to provider
			sendSyncExchange(provider, exchange, context);

			// send OUT or Fault, error to the monitor if Response way is
			// defined
			sendResponseCopyIfNeeded(monitor, exchange, context);

		} catch (JBIException e) {
			logger.log(Level.SEVERE, e.getMessage(), e);
			try {
				exchange.setFault(e);
			} catch (MessagingException ex) {
				logger.severe(ex.getMessage());
				exchange.setError(ex);
			}
		}
	}

	/**
	 * 
	 * @param monitor
	 * @param exchange
	 * @param context
	 * @throws PEtALSCDKException
	 * @throws JBIException
	 */
	protected void sendRequestCopyIfNeeded(Consumes monitor, Exchange exchange,
			ExchangeContext context) throws PEtALSCDKException, JBIException {
		String way = context.getExtensions().getValue(WAY, REQUESTWAY);

		if (way.toLowerCase().indexOf(REQUESTWAY) >= 0) {
			NormalizedMessage inMessage = exchange.getInMessage();

			Exchange monitorExchange = PatternHelper.createExchange(monitor,
					MEPConstants.IN_ONLY_PATTERN.value(), inMessage, context);
			logger.info(
					"send a copy of the IN message to ...\n - itf="
							+ monitor.getInterfaceName() + "\n - srv="
							+ monitor.getServiceName() + "\n - ep ="
							+ monitor.getEndpointName());
			context.send(monitorExchange);
		}

	}

	/**
	 * Send the exchange to the provider (only set the information, the
	 * framework will send it lately)
	 *
	 * @param provider
	 * @param exchange
	 * @param context
	 * @throws PEtALSCDKException
	 * @throws JBIException
	 */
	protected void sendSyncExchange(Consumes provider,
			Exchange originalExchange, ExchangeContext context)
			throws PEtALSCDKException, JBIException {
		NormalizedMessage inMessage = originalExchange.getInMessage();

		Exchange providerExchange = PatternHelper.createExchange(provider,
				originalExchange.getExchangePattern(), inMessage, context);

		logger.info(
				"send the exchange to ...\n - itf=" + provider.getInterfaceName()
						+ "\n - srv=" + provider.getServiceName() + "\n - ep ="
						+ provider.getEndpointName());
		context.sendSync(providerExchange);

		// process the response
		NormalizedMessage receivedMessage = PatternHelper.processSentExchange(
				providerExchange, originalExchange, context);
		if (receivedMessage != null) {
			PatternHelper.copy(receivedMessage, originalExchange
					.getOutMessage());
		}
	}

	/**
	 * 
	 * @param monitor
	 * @param exchange
	 * @param context
	 * @throws PEtALSCDKException
	 * @throws JBIException
	 */
	protected void sendResponseCopyIfNeeded(Consumes monitor,
			Exchange exchange, ExchangeContext context)
			throws PEtALSCDKException, JBIException {
		String way = context.getExtensions().getValue(WAY, REQUESTWAY);

		if (way.toLowerCase().indexOf(RESPONSEWAY) >= 0
				&& !exchange.isDoneStatus()) {
			Exchange monitorExchange = context.createConsumeExchange(monitor,
					MEPConstants.IN_ONLY_PATTERN.value());
			if (exchange.isErrorStatus()) {
				monitorExchange.setError(exchange.getError());
			} else if (exchange.isActiveStatus()) {
				if (exchange.getFault() != null) {
					Fault monitorFault = monitorExchange.createFault();
					PatternHelper.copy(exchange.getFault(), monitorFault);
				} else {
					PatternHelper.copy(exchange.getOutMessage(),
							monitorExchange.getInMessage());
				}
			}
			context.send(monitorExchange);
			logger.info(
					"send a copy of the OUT/Fault/error message to ...\n - itf="
							+ monitor.getInterfaceName() + "\n - srv="
							+ monitor.getServiceName() + "\n - ep ="
							+ monitor.getEndpointName());
		}
	}
	
	/*
	 * (non-Javadoc)
	 * @see org.ow2.petals.se.eip.patterns.Pattern#init()
	 */
	public void init() {
	}

	/*
	 * (non-Javadoc)
	 * @see org.ow2.petals.se.eip.patterns.AbstractPattern#getPatternName()
	 */
	@Override
	protected String getPatternName() {
		return "WireTap";
	}

	/*
	 * (non-Javadoc)
	 * @see org.ow2.petals.se.eip.patterns.AbstractPattern#validateMEP(java.net.URI)
	 */
	@Override
	protected boolean validateMEP(URI mep) {
		return true;
	}
}
